{
  Unit :      VisQuery
  Component : TVisQuery
  Author :    Milos Dragovic
  Created:    15.12.1998.

 This component is distributed as Freeware. There is no charge or warranty
 whatsoever. The author will not be held responsible for any direct or
 indirect damage resulting from the use of the component or any
 derivatives. You are allowed to include the component code in
 any commercial or non commercial product under the condition that
 you mention me in that software. You are also allowed to
 make as many copies of the component as you want and distribute them,
 as long as you do not receive any money for the copying/distribution.

 You can send suggestions, bug reports and improvements
     to dragomil@EUnet.yu

   Requirements: Delphi 3 or higher

Installation:

    - In short, perform the Standard COmponent INstallation procedure.
         (SCOIN - my small contribution to the world of abbrevations)
    - In detail, (if you don't know what the #$%@#$ I'm saying here),
  I suggest that you create a new package (in Delphi: File/New/Package),
  save it as, say, "DownloadedComponents.dpk", click on the "Add" button,
  browse and find this unit (VisQuery.pas), select it, Click on
  "Install" and the TVisQuery component should be on Delphi's
  "DataAccess" page. Naturally, you can use the same package later for
  installing some other components. That's it.

 Description:

 TVisQuery component is a TQuery descendant capable of creating and
 showing a simple form with a DBGrid, DBNavigator and a specified
 CoolBar with whatever controls you put on it. Grid and navigator
 are (of course) linked to the query. Rearranging columns of the grid
 changes the ORDER BY part of the sql statement so
 that the query result is sorted in order of the fields in the grid. Also,
 search is supported using CTRL+F (Find) and F3 (find again).
 If you wish, TVisQuery will store the position and size of it's
 autocreated form in a file. VisQuery can refresh the query each time
 it's form is re-activated.

 Introduced methods, properties and an event:

- procedure Show

     When called for the first time this procedure will create the
     form and show it, later it will simply show the existing form.

- procedure Hide

     Hides the VisQuery's form

- procedure OpenQuery

     (Since Open cannot be overriden). This is same as calling Open
     except that it will show you the SQL string before the error message.
     It's useful if you are changing - constructing the SQL statement
     at runtime. Also, if SQLHourglass is set to true it will change
     the mouse pointer for the duration of the Open method.
     Of course, you can still use Open if you don't need these features.

- property Source

     Since it is needed for connecting to the grid and navigator
     components, TVisQuery creates a TDataSource component.
     Source is that Datasource - you can use it
     instead of making a separate one. (Accessible only at run-time)

- properties Caption, Icon, Font and Menu

      Used for the autocreated form's caption, icon, font and, yes you
      guessed it, it's menu.

- property KeepSettings : boolean

      If true, autocreated form's settings (position and size) will
      be saved and loaded from the "VQFile.cfg" file which will be
      searched and saved in the application's directory.

- property ExtraCoolBar : TCoolBar

      If you want the autocreated form to contain some extra controls
      you can put them on a coolbar (put the coolbar on some temporary
      form - or on a normal form and set visible to false). When the
      VisQuery's form is created, that coolbar will be added to it (and
      therefore removed from the form you placed it on). This way you
      can customize the autocreated form by adding toolbars, combo boxes
      for filtering and so on.

- property RefreshOnActivate : boolean

       If true, every time the VisQuery's form is activated the query
       will be closed and reopened.

- property ShowSQLHourglass : boolean

       If true, OpenQuery method will change the mouse pointer into
       crSQLWait for the duration of the Open method.

- event AfterShow : TDataSetNotifyEvent

      This event is called whenever the show method is called or the
      form is activated - received focus.
}

unit VisQuery;

interface
{$R VisQuery.res}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, DBTables, ComCtrls, DBCtrls, StdCtrls, Buttons,
  DBGrids, Menus;

type

 TVQFormSettings = record
     Name : string[50];
     Maximized : boolean;
     Height, Width, Left, Top : integer;
     end;

 TVisQuery = class(TQuery)
  private
  protected
    FAfterShow : TDAtaSetNotifyEvent;
    FCaption : string;
    FSource : TDataSource;
    FCoolBar : TCoolBar;
    FIcon : TIcon;
    FGrid : TDBGrid;
    FNAvigator : TDBNAvigator;
    FFont : TFont;
    FSQLHourglass, FRefresh, FKeepSettings : boolean;
    FMenu : TMainMenu;
    procedure SetIcon (i : TIcon);
    procedure SetFont (F : TFont);
    procedure SetMenu (m : TMainMenu);
    procedure FormActivate (sender : TObject);
  protected
    SearchValue : string;
    SearchField : TField;
    procedure Notification (AComponent: TComponent; Operation: TOperation); override;
    procedure CreateForm; virtual;
    procedure MovedCol(Sender: TObject; FromIndex, ToIndex: Longint); virtual;
    procedure LoadSettings; virtual;
    procedure SaveSettings; virtual;
    procedure GridKeyDown (sender : TObject; var Key: Word; Shift: TShiftState);
  public
    Form : TForm;
    constructor Create (Owner : TComponent); override;
    destructor Destroy; override;
    property Source : TDataSource read FSource;
    procedure Show; virtual;
    procedure Hide;
    procedure OpenQuery; virtual;
  published
    property AfterShow : TDAtaSetNotifyEvent read FAfterShow write FAfterShow;
    property Caption : string read FCaption write FCaption;
    property ExtraCoolBar : TCoolBar read FCoolBar write FCoolBar;
    property Icon : TIcon read FICon write SetIcon;
    property Menu : TMainMenu read FMenu write SetMenu;
    property Font : TFont read FFont write SetFont;
    property KeepSettings : boolean read FKeepSettings write FKeepSettings;
    property RefreshOnActivate : boolean read FRefresh write FRefresh;
    property ShowSQLHourGlass : boolean read FSQLHourglass write FSQLHourglass;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('DataAccess', [TVisQuery]);
end;

constructor TVisQuery.Create (Owner : TComponent);
begin
     inherited Create(owner);
     FIcon := TIcon.create;
     FFont := TFont.create;
     if csdesigning in componentstate then exit;
     FSource := TDataSource.create(self);
     FSource.dataset := self;
end;
destructor TVisQuery.Destroy;
begin
   if FKeepSettings and assigned(Form) then SaveSettings;
   if assigned (FSource) then FSource.free;
   FICon.free;
   FFont.free;
   inherited destroy;
end;
procedure TVisQuery.Notification (AComponent: TComponent; Operation: TOperation);
begin
  inherited notification (AComponent,Operation);
  if Operation = OpRemove then
   begin
     if AComponent = FCoolBar then FCoolBar := nil;
     if AComponent = FMenu then FMenu := nil;
   end;
end;
procedure TVisQuery.SetIcon (i : TIcon);
begin
   FIcon.Assign(i);
   if assigned(Form) then Form.Icon := i;
end;
procedure TVisQuery.Show;
begin
  if not active then OpenQuery;
  if Form = nil then CreateForm
  else Form.show;
end;
procedure TVisQuery.CreateForm;
begin
  Form := TForm.create(self);
  Form.caption := caption;
  Form.icon := FIcon;
  if assigned(Font) then Form.font := FFont;
  FGrid := TDBGrid.create(Form);
  FGrid.parent := Form;
  FGrid.align := alclient;
  FGrid.DataSource := FSource;
  FGrid.OnColumnMoved := MovedCol;
  FGrid.OnKeyDown := GridKeyDown;
  FNavigator := TDBNAvigator.create(Form);
  FNavigator.parent := Form;
  FNavigator.align := albottom;
  FNavigator.DataSource := FSource;
  FNavigator.VisibleButtons := [nbfirst, nbprior, nbnext, nblast];
  if FKeepSettings then LoadSettings;
  if assigned(FCoolBar) then
    begin
     FCoolBar.parent := Form;
     FCoolBar.Visible := true;
    end;
  if assigned(FMenu) then
    begin
      if FMenu.Handle = 0 then
         Form.Menu := FMenu
      else
        begin
          Form.Menu := TMAinMenu.create(Form);
          Form.Menu.Merge (FMenu);
        end;
    end;
  MovedCol(self, -1, -1);
  Form.Show;
  Form.OnActivate := FormActivate;
end;

procedure TVisQuery.SetFont (F : TFont);
begin
   FFont.assign(f);
   if Form <> nil then Form.font := f;
end;
procedure TVisQuery.SetMenu (m : TMainMenu);
begin
   FMenu := m;
   if Form <> nil then Form.Menu := m;
end;

procedure TVisQuery.GridKeyDown (sender : TObject; var Key: Word; Shift: TShiftState);
var ExRec : longint;
begin
  if key = VK_F3 then
    begin
      if (SearchField = nil) or (SearchValue='') then exit;
      Exrec := RecNo;
      DisableControls;
      next;
{      if eof then
        begin
          if MessageDlg ('End of query reached. Do you want to start from the top?', mtConfirmation, [mbYes, mbNo], 0) = mrYes
            then First
          else
            begin
              RecNo := ExRec;
              EnableControls;
              exit;
            end;
        end;}
      while not (eof) and (SearchField.asstring <> SearchValue) do
         next;
      if eof and (SearchField.asstring <> SearchValue) then
       begin
         ShowMessage('Value "'+SearchValue+'" was not found in field '+SearchField.DisplayLabel);
         RecNo := Exrec;
       end;
      FGrid.SelectedField := SearchField; 
      EnableControls;
    end;
  if (key <> ord('F')) or not (ssctrl in Shift) then exit;
  SearchField := FGrid.SelectedField;
  SearchValue := InputBox ('Find in field : '+SearchField.DisplayLabel, 'Enter value', '' );
  if SearchValue = '' then exit;
  Exrec := RecNo;
  DisableControls;
  First;
  while not (eof) and (SearchField.asstring <> SearchValue) do
     next;
  if SearchField.asstring <> SearchValue then
    begin
      ShowMessage('Value "'+SearchValue+'" was not found in field '+SearchField.DisplayLabel);
      RecNo := Exrec;
    end;
  EnableControls;
end;
procedure TVisQuery.MovedCol(Sender: TObject; FromIndex, ToIndex: Longint);
var orderby, orderName : string;i : integer;
procedure RemoveOrder;
var query : string;p : integer;i, c : integer;
begin
  query := lowercase(sql.Text);
  sql.text := '';
  c := 0;
  while query<>'' do
    begin
      p := pos ('order', query);
      if p = 0 then
        begin
          sql.text := sql.text + query;
          exit;
        end;
      for i := 1 to p do
        if query = '"' then c := c + 1;
      if trunc(c/2)*2 <> c/2*2 then
        begin
          sql.text := sql.text + copy(query, 1, p+5);
          query := copy (query, p+6, 5555);
        end
      else
        begin
          sql.text := sql.text + copy(query, 1, p-1);
          exit;
        end;
    end;
end;
begin
  close;
  RemoveOrder;
  orderby := ' ORDER BY ';
  for i := 0 to fieldcount-1 do
    begin
      if fields[i].FieldKind = fkdata
       then OrderName := fields[i].FieldName
       else if fields[i].FieldKind = fklookup
         then OrderName := fields[i].KeyFields;
      if orderby <>' ORDER BY ' then orderby := orderby + ', ';
      orderby := orderby + OrderName;
    end;
  sql.add(OrderBy);
  OpenQuery;
  if FromIndex <> -1 then abort;
end;

procedure TVisQuery.OpenQuery;
var OldCursor : TCursor;
begin
   OldCursor := screen.Cursor;
   if FSQLHourglass then screen.Cursor := crSQLWait;
   try
      open;
   except
      ShowMessage('Query failed : '+sql.text);
      raise;
   end;
   screen.Cursor := OldCursor;
end;

procedure TVisQuery.LoadSettings;
var rec :TVQFormSettings;f : TFileStream;
begin
   if not fileexists (ExtractFileDir(application.ExeName)+'\VQFile.cfg')
      then exit;
   f := TFileStream.create(ExtractFileDir(application.ExeName)+'\VQFile.cfg', fmopenread);
   while f.position < f.size-1 do
      begin
        f.read (rec, sizeof(rec));
        if rec.Name = name then
          begin
            Form.Width := rec.Width;
            Form.Height := rec.Height;
            Form.left := rec.left;
            Form.top := rec.top;
            if rec.Maximized then  Form.WindowState := wsmaximized;
            break;
          end
      end;
   f.free;
end;

procedure TVisQuery.SaveSettings;
var f : TFileStream;rec :TVQFormSettings;
begin
   if fileexists (ExtractFileDir(application.ExeName)+'\VQFile.cfg')
     then f := TFileStream.create(ExtractFileDir(application.ExeName)+'\VQFile.cfg', fmopenreadwrite)
     else f := TFileStream.create(ExtractFileDir(application.ExeName)+'\VQFile.cfg', fmcreate);
   while f.position < f.size-1 do
      begin
        f.read (rec, sizeof(rec));
        if rec.Name = name then
            begin
              f.position := f.position - sizeof(rec);
              break;
            end;
      end;
   rec.name := Name;
   rec.Width := Form.Width;
   rec.Height := Form.Height;
   rec.Maximized := Form.WindowState = wsmaximized;
   rec.left := Form.left;
   rec.top := Form.Top;
   f.write(rec, sizeof(rec));
   f.free;
end;
procedure TVisQuery.Hide;
begin
  if assigned(Form) then Form.hide;
end;
procedure TVisQuery.FormActivate (sender : TObject);
begin
  if Application.terminated then exit;
  if assigned (FAfterShow) then
     FAfterShow (self);
  if FRefresh then
    begin
      close;
      OpenQuery;
    end;
end;
end.
